/* Copyright (c) 2006-2015 Polybus Systems Corp., Westford, MA, USA */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "sys_basher.h"


long fpDoubleTest(void *base,SInt64 limit,long num,long num_threads,long firsttime,
		  double exp[MAX_SIZES][NUM_FP_TESTS],result fp_rslt[NUM_FP_TESTS][MAX_SIZES],SInt32 vnum,SInt64 time_scale[MAX_SIZES][NUM_FP_TESTS])
{
  SInt64 i,j,length,max;
  long num_sizes,errors=0,tnum;
  double bwidth[64][8],got[64][8];
  double *fdata;
  char string[1024];
  char *test_names[] = {
    "Add RR",
    "Add MR",
    "Mpy RR",
    "Mpy MR",
    "Div RR",
    "Div MR",
    "SofP MRR",
    "SofP MMR",
    0L};
  char *results[16];
  static char *flops="FLOPS";
  SInt64 src[] = {0x3FE278E2EF9D6B9ELL,
		  0x3FE278E2EF9D6B9ELL,
		  0xBFE278E2EF9D6B9DLL,
		  0xBFE278E2EF9D6B9DLL,
		  0x3E06C45A0A881B87LL,
		  0x3E06C45A0A881B87LL,
		  0x41D67D1AD7C067B1LL,
		  0x41D67D1AD7C067B1LL};

		     

  fdata = (double *)src;

  max = limit & 0xFFFFFFFFFFFF0000LL;
  if(max > 0x20000000LL) max = 0x20000000LL;
  max /= sizeof(double);
  sprintf(string,"\nThread %ld of %ld: Running Double Precision FP Tests\n",num,(long)num_threads);
  printTranscript(string,TRUE);
  
  length = 0x400L;
  num_sizes = 0;
  while(length < max)
    {
      tnum = 0;

      fillDoubleMem(base,max,fdata,4L);

      got[num_sizes][tnum] = rrDoubleSum(max,length,fdata,&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrDoubleSum(base,max,length,fdata,&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;
 
      fillDoubleMem(base,max,&fdata[4],4L);

      got[num_sizes][tnum] = rrDoubleMpy(max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrDoubleMpy(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = rrDoubleDiv(max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrDoubleDiv(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      fillDoubleMem(base,max,&fdata[4],1L);
      got[num_sizes][tnum] = mrrDoubleSumProd(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mmrDoubleSumProd(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkDouble(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      sprintf(fp_rslt[tnum][num_sizes].name,"Double FP %s",test_names[tnum]);
      fp_rslt[tnum][num_sizes].size = length;

      num_sizes++;
      length = length << 2;
    }

  length = 0x400L;
  errors = 0;
  sprintf(string,"\nThread %ld of %ld: Double Precision FP\n",num,(long)num_threads);
  printTranscript(string,FALSE);
  displayTitles("Array Size",test_names,8L);
  for(i=0;i<8;i++) results[i] = flops;
  displayTitles("",results,8L);
  for(i=0;i<num_sizes;i++)
    {
      iscale(string,"B",length);
      displayLine(string,"/S",bwidth[i],8L);
      length = length << 2;
    }
  printTranscript("\n",FALSE);
  if(firsttime)
    {
      for(i=0;i<num_sizes;i++)
	{
	  for(j=0;j<NUM_FP_TESTS;j++)
	    {
	      exp[i][j] = got[i][j];
	    }
	}
    }
  else
    {
      length = 0x400L;
      for(i=0;i<NUM_FP_TESTS;i++) results[i] = "PASS/FAIL";
      displayTitles("Array Size",test_names,8L);
      displayTitles("",results,8L);
      for(i=0;i<num_sizes;i++)
	{
	  iscale(string,"B",length);
	  errors += displayPassFail(string,got[i],exp[i],NUM_FP_TESTS);
	  length = length << 2;
	}
    }


  sprintf(string,"Thread %ld, Floating Point Test",num);
  if(errors == 0)
    {
      sprintf(string,"PASSED\tThread %ld of %ld: Double Precision Tests\n",(long)num,(long)num_threads);
      printTranscript(string,TRUE);
      return(1L);
    }
  else
    {
      sprintf(string,"FAILED\tThread %ld of %ld: Double Precision Tests\n",(long)num,(long)num_threads);
      printTranscript(string,TRUE);
      return(0L);
    }

}


void fillDoubleMem(double * base,SInt64 max,double ivalue[],long words)
{
  double *indx;
  SInt64 i,j;

  indx = base;

  for(i=0;i<max;i+=words)
    {
      for(j=0;j<words;j++)
	{
	  *indx++ = ivalue[j];
	}
    }
}
double rrDoubleSum(SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,w=0,x=0,y=0,z=0,start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  while(!done)
    {
      acc = 0;
      w = fdata[0];
      x = fdata[1];
      y = fdata[2];
      z = fdata[3];
      start = getTime();
      for(k=0;k<scale;k++)
	{


	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  w += x;
		  x += y;
		  y += z;
		  x += w;
		}
	    }
	}
      end = getTime();
      acc = w + x + y +z;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;
  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double mrDoubleSum(double * base,SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,w=0,x=0,y=0,z=0,start=0,end=0,usecs=0,ops=0,*indx=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);


  while(!done)
    {
      w = 0;
      x = 0;
      y = 0;
      z = 0;
      acc = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  w += *indx++;
		  x += *indx++;
		  y += *indx++;
		  x += *indx++;
		}
	    }
	}
      end = getTime();
      acc = w + x + y +z;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double rrDoubleMpy(SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,x=0,y=0,start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);



  while(!done)
    {
      acc = 0;
      start = getTime();
      x = 1.000000001;
      y = -1.0000000001;

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  x *= fdata[0];
		  y *= fdata[1];
		  x *= fdata[2];
		  y *= fdata[3];
		}
	    }
	}
      end = getTime();
      acc = x + y;

      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double mrDoubleMpy(double * base,SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,x=0,y=0,start=0,end=0,usecs=0,ops=0,*indx=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);



  while(!done)
    {
      acc = 0;
      x = 1.000000001;
      y = -1.0000000001;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  x *= *indx++;
		  y *= *indx++;
		  x *= *indx++;
		  y *= *indx++;
		}
	    }
	}
      end = getTime();
      acc = x + y;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double rrDoubleDiv(SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,x=0,y=0,start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);



  while(!done)
    {
      acc = 0;
      x = 0x11111111111111LL;
      y = 0xABCDEF01234567LL;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  x /= fdata[0];
		  y /= fdata[1];
		  x /= fdata[2];
		  y /= fdata[3];
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double mrDoubleDiv(double * base,SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,start=0,x=0,y=0,end=0,usecs=0,ops=0,*indx=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  acc = 1.0;

  while(!done)
    {
      acc = 0;
      x = 0x11111111111111LL;
      y = 0xABCDEF01234567LL;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  x /= *indx++;
		  y /= *indx++;
		  x /= *indx++;
		  y /= *indx++;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double mrrDoubleSumProd(double * base,SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,sx=0,sy=0,x=0,y=0,start=0,end=0,usecs=0,ops=0,*indx;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);


  acc = 0;

  while(!done)
    {
      acc = 0;
      x = 0;
      y = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  indx = base;
	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=8)
		{
		  sx = *indx++;
		  sx *= fdata[0];
		  x += sx;
		  sy = *indx++;
		  sy *= fdata[0];
		  y += sy;
		  sx = fdata[0];
		  sx *= *indx++;
		  x += sx;
		  sy = fdata[0];
		  sy *= *indx++;
		  y += sy;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

double mmrDoubleSumProd(double * base,SInt64 max,SInt64 length,double fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  double acc=0,start=0,x=0,y=0,sx=0,sy=0,end=0,usecs=0,ops=0,*indx=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);


  acc = 0;

  while(!done)
    {
      acc = 0;
      x = 0;
      y = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  indx = base;
	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=8)
		{
		  sx = *indx++;
		  sx *= *indx++;
		  x += sx;
		  sy = *indx++;
		  sy *= *indx++;
		  y += sy;
		  sx = *indx++;
		  sx *= *indx++;
		  x += sx;
		  sy = *indx++;
		  sy *= *indx++;
		  y += sy;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

long fpFloatTest(void *base,SInt64 limit,long num,long num_threads,long firsttime,
		 float exp[MAX_SIZES][NUM_FP_TESTS],result fp_rslt[NUM_FP_TESTS][MAX_SIZES],SInt32 vnum,SInt64 time_scale[MAX_SIZES][NUM_FP_TESTS])
{
  SInt64 i,j,length,max;
  long num_sizes,errors=0,tnum;
  double bwidth[64][8];
  float got[64][8];
  float *fdata;
  char string[1024];
  char *test_names[] = {
    "Add RR",
    "Add MR",
    "Mpy RR",
    "Mpy MR",
    "Div RR",
    "Div MR",
    "SofP MRR",
    "SofP MMR",
    0L};
  char *results[16];
  static char *flops="FLOPS";

  unsigned long src[] = {0x40490FDBL,
			 0x40490FDBL,
			 0xC0490FD7L,
			 0xC0490FD7L,
			 0x402DF854L,
			 0x402DF854L,
			 0x3EBC5AB2L,
			 0x3EBC5AB2L,
			 0x402DF854L,
			 0x3EBC5AB2L,
			 0x402DF854L,
			 0x3EBC5AB2L
  };

  max = limit & 0xFFFFFFFFFFFF0000LL;
  if(max > 0x40000000LL) max = 0x40000000LL;
  max /= sizeof(float);
  
  sprintf(string,"Thread %ld of %ld: Running Single Precision FP Tests\n",(long)vnum,(long)num_threads);
  printTranscript(string,TRUE);
  length = 0x400L;
  num_sizes = 0;
  fdata = (float *)src;


  while(length < max)
    {
      tnum = 0;


      fillFloatMem(base,max,fdata,4l);
      got[num_sizes][tnum] = rrFloatSum(max,length,fdata,&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrFloatSum(base,max,length,fdata,&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;
 
      fillFloatMem(base,max,&fdata[4],4l);

      got[num_sizes][tnum] = rrFloatMpy(max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrFloatMpy(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = rrFloatDiv(max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mrFloatDiv(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      fillFloatMem(base,max,&fdata[4],1L);
      got[num_sizes][tnum] = mrrFloatSumProd(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);
      tnum++;

      got[num_sizes][tnum] = mmrFloatSumProd(base,max,length,&fdata[4],&bwidth[num_sizes][tnum],&fp_rslt[tnum][num_sizes],&time_scale[num_sizes][tnum]);
      chkFloat(firsttime,exp[num_sizes][tnum],got[num_sizes][tnum],&fp_rslt[tnum][num_sizes]);
      fp_rslt[tnum][num_sizes].size = length;
      sprintf(fp_rslt[tnum][num_sizes].name,"Single FP %s",test_names[tnum]);

      num_sizes++;
      length = length << 2;
    }

  length = 0x400L;
  errors = 0;
  sprintf(string,"\nThread %ld of %ld: Single Precision FP Performance\n",(long)vnum,(long)num_threads);
  printTranscript(string,FALSE);
  displayTitles("Array Size",test_names,8L);
  for(i=0;i<8;i++) results[i] = flops;
  displayTitles("",results,8L);
  for(i=0;i<num_sizes;i++) 
    {
      iscale(string,"B",length);
      displayLine(string,"/S",bwidth[i],8L);
      length = length << 2;
    }

  errors = 0;
  length = 0x400L;
  for(i=0;i<8;i++) results[i] = "PASS/FAIL";
  printTranscript("\n",FALSE);

  if(firsttime)
    {
      for(i=0;i<num_sizes;i++)
	{
	  for(j=0;j<NUM_FP_TESTS;j++)
	    {
	      exp[i][j] = got[i][j];
	    }
	}
    }
  else
    {
      length = 0x400L;
      for(i=0;i<NUM_FP_TESTS;i++) results[i] = "PASS/FAIL";
      displayTitles("Array Size",test_names,8L);
      displayTitles("",results,8L);
      for(i=0;i<num_sizes;i++)
	{
	  iscale(string,"B",length);
	  errors += displayFloatPassFail(string,got[i],exp[i],NUM_FP_TESTS);
	  length = length << 2;
	}
    }

  
  sprintf(string,"Thread %ld, Floating Point Test",num);
  if(errors == 0)
    {
      sprintf(string,"PASSED\tThread %ld of %ld: Single Precision Tests\n",(long)vnum,(long)num_threads);
      printTranscript(string,TRUE);
      return(1L);
    }
  else
    {
      sprintf(string,"FAILED\tThread %ld of %ld: Single Precision Tests\n",(long)vnum,(long)num_threads);
      printTranscript(string,TRUE);
      return(0L);
    }

}


void fillFloatMem(float *base,SInt64 max,float ivalue[],long words)
{
  float *indx;
  SInt64 i,j;

  indx = base;
  for(i=0;i<max;i+=words)
    {
      for(j=0;j<words;j++)
	{
	  *indx++ = ivalue[j];
	}
    }
}
float rrFloatSum(SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc=0,w=0,x=0,y=0,z=0;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  while(!done)
    {
      w = fdata[0];
      x = fdata[1];
      y = fdata[2];
      z = fdata[3];
      acc = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{


	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  w += x;
		  x += y;
		  y += z;
		  x += w;
		}
	    }
	}
      end = getTime();
      acc = w + x + y +z;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float mrFloatSum(float * base,SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc=0,w=0,x=0,y=0,z=0,*indx=0;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);


  while(!done)
    {
      acc = 0;
      w = 0;
      x = 0;
      y = 0;
      z = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  w += *indx++;
		  x += *indx++;
		  y += *indx++;
		  x += *indx++;
		}
	    }
	}
      end = getTime();
      acc = w + x + y +z;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float rrFloatMpy(SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc=0,x=0,y=0;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);



  while(!done)
    {
      acc = 0;
      x = -3.141592628;
      y = x * x;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  x *= fdata[0];
		  y *= fdata[1];
		  x *= fdata[2];
		  y *= fdata[3];
		}
	    }
	}
      end = getTime();
      acc = x + y;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float mrFloatMpy(float * base,SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc=0,x=0,y=0,*indx=0;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);



  while(!done)
    {
      acc = 0;
      x = -3.141592628;
      y = x * x;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  x *= *indx++;
		  y *= *indx++;
		  x *= *indx++;
		  y *= *indx++;
		}
	    }
	}
      end = getTime();
      acc = x + y;
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float rrFloatDiv(SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc=0,x=0,y=0;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  acc = 1.0;

  while(!done)
    {
      acc = 0;
      start = getTime();
      x = -12345677;
      y = 0x89ABCDEF;

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      for(j=0;j<length;j+=4)
		{
		  x /= fdata[0];
		  y /= fdata[1];
		  x /= fdata[2];
		  y /= fdata[3];
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float mrFloatDiv(float * base,SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;

  float acc=0,x=0,y=0,*indx=0;
  double start=0,end=0,usecs=0,ops=0;

  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  acc = 1.0;

  while(!done)
    {
      acc = 0;
      x = -12345677;
      y = 0x89ABCDEF;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=4)
		{
		  x /= *indx++;
		  y /= *indx++;
		  x /= *indx++;
		  y /= *indx++;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float mrrFloatSumProd(float * base,SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc,x,y,sx,sy,*indx;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);


  acc = 0;

  while(!done)
    {
      acc = 0;
      x = 0;
      y = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  indx = base;
	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=8)
		{
		  sx = *indx++;
		  sx *= fdata[0];
		  x += sx;
		  sy = fdata[0];
		  sy *= *indx++;
		  y += sy;
		  sx = *indx++;
		  sx *= fdata[0];
		  x += sx;
		  sy = fdata[0];
		  sy *= *indx++;
		  y += sy;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

float mmrFloatSumProd(float * base,SInt64 max,SInt64 length,float fdata[],double *bwidth,result *rslt,SInt64 *time_scale)
{
  float acc,x,y,sx,sy,*indx;
  double start=0,end=0,usecs=0,ops=0;
  SInt64 i,j,k,loops,scale = *time_scale;
  Boolean firstTime=0,done = 0;


  if(scale == 0)
    {
      scale = 1;
      firstTime = TRUE;
    }
  loops = numLoops(length);

  acc = 0;

  while(!done)
    {
      acc = 0;
      x = 0;
      y = 0;
      start = getTime();

      for(k=0;k<scale;k++)
	{

	  indx = base;
	  for(i=0;i<loops;i++)
	    {
	      indx = base;
	      for(j=0;j<length;j+=8)
		{
		  sx = *indx++;
		  sx *= *indx++;
		  x += sx;
		  sy = *indx++;
		  sy *= *indx++;
		  y += sy;
		  sx = *indx++;
		  sx *= *indx++;
		  x += sx;
		  sy = *indx++;
		  sy *= *indx++;
		  y += sy;
		}
	    }
	}
      end = getTime();
      usecs = end - start;
      if(firstTime)
	{
	  scale = scale * 2LL;
	  *time_scale = scale;
	}
      done = !firstTime;
      if(usecs > 250000.0) firstTime = 0;
	
    }
  scale = scale/2LL;

  if(firstTime)
    {
      *time_scale = scale;
    }

  acc = x + y;
  ops = loops * scale;
  ops *= length;
  *bwidth = ops/usecs;
  updateBandwidth(*bwidth,usecs,ops,rslt);
  return(acc);
}

SInt64 numLoops(SInt64 length)
{
  SInt64 loops;
  loops = 0x8000000LL/length;
  if(loops == 0) loops = 1LL;
  return(loops);
}
